<?php

# This file should contain all the general helper functions,
# it is included by prolog.php, but can also be included before
# if necessary. Note that prolog.php will not include it a
# second time if it was already included.

# Returns the location of the video out sysfs, with a trailing slash
function sysfs_vout_device() {
     if (file_exists('/sys/class/x-display/disp0/device')) {
	  return '/sys/class/x-display/disp0/device/';
     } else {
	  return '/sys/class/graphics/fb0/device/';
     }
}

# This function returns a multidimensional array. Each IP
# address gets a row, and each row has three entries: the IP
# address, prefix length, interface type (e.g., 'ethernet',
# 'modem') and the interface name (e.g., 'eth0', 'ppp0'). The "main"
# address is returned in the first row. The loopback interface
# is not returned.
function get_ip_addrs() {
    $ipcmd=shell_exec('ip -o -f inet addr show');
    preg_match_all("@^\d+:\s*(\S+)\s+.*\binet\s+([\d.]+)(?:/(\d+)|\s+)@m", $ipcmd, $matches, PREG_SET_ORDER);

    $addrs = array();
    foreach ($matches as $idx => $value) {
        $iface = $value[1];
        $addr = $value[2];
        $plen = $value[3];

        if ( ! $plen )
            $plen = 32;

        $order = 0; // to order the addresses by decreasing importance

        // add n in [0-99] to ordering to keep ordering if all other
        // ordering criteria are equal
        $order += $idx;

        if ( strncmp($iface, 'eth', 3) == 0 )
            $type = 'ethernet';
        elseif ( strncmp($iface, 'ppp', 3) == 0 )
            $type = 'modem';
        elseif ( $iface == 'lo' )
            $type = 'loopback';
        if ( $iface == 'lo' )
            continue;

        // add n*100 to order for interface type (n in [0,9])
        if ( $type == 'modem' )
            $order += 100;
        elseif ( $type == 'ethernet' )
            $order += 800;
        else
            $order += 900;

        // add n*1000 to order for address class (n in [0,9])
        if (strncmp($addr, '169.254', 7) == 0 )
            $order += 8000; // link-local
        if (strncmp($addr, '127.', 4) == 0 )
            $order += 9000; // host-local

        $addrs[$order] = array($addr, $plen, $type, $iface);
    }

    // apply the ordering
    ksort($addrs);

    return $addrs;
}

function getMAC(){
	$info=`ifconfig eth0`;
	$macp=strpos($info,"HWaddr ");
	if ($macp===false) return "";
	$macp=$macp+strlen("HWaddr ");
	$mac = substr($info, $macp, strpos($info," ",$macp)-$macp);
	return $mac;
}

function net_prefix_to_mask($length)
{
    $nm = '';
    $ne = 0;
    while ($length >= 8) {
        $nm .= ($ne ? '.255' : '255');
        $length -=8;
        $ne++;
    }
    while ($ne < 4) {
        if ($ne)
            $nm .= '.';
        $nm .= (0xFF << (8 - $length) & 0xFF);
        $ne++;
        $length = 0;
    }
    return $nm;
}

function net_mask_to_prefix($netmask)
{
    $length = 0;
    $vals = explode('.', $netmask);
    if ( count($vals) != 4 )
        return false;
    $expect_zero = false;
    foreach ($vals as $val) {
        if ( ! is_numeric($val) || $val < 0 || $val > 255 )
            return false;
        if ($expect_zero) {
            if ($val != 0)
                return false;
            else
                continue;
        }
        switch ($val) {
        case 255:
            $length += 8; break;
        case 254:
            $length += 7; break;
        case 252:
            $length += 6; break;
        case 248:
            $length += 5; break;
        case 240:
            $length += 4; break;
        case 224:
            $length += 3; break;
        case 192:
            $length += 2; break;
        case 128:
            $length += 1; break;
        case 0:
            $length += 0; break;
        default:
            return false;
        }
        if ($val != 255)
            $expect_zero = true;
    }
    return $length;
}

function ip_check($ip) {
      $vals=explode (".",$ip);
      if (count($vals)!=4) return false;
      foreach ($vals as $val) {
	   if (!is_numeric($val) || $val<0 || $val>255) return false;
      }
      if ($vals[0]==0) return false;
      if ($vals[0]==255 && $vals[1]==255 && $vals[2]==255 && $vals[3]==255) return false;
      
      return true;
}
function netmask_check($ip) {
    $len = net_mask_to_prefix($ip);
    return ( $len !== false && $len > 0 && $len < 32 );
}
function ipconf_check($adr, $msk, $gateway) {
	$a=ip2long($adr);
	$m=ip2long($msk);
	$g=ip2long($gateway);

	if ( ($a & $m) == $a ) return false;
	if ( ($a & $m) != ($g & $m) ) return false;
	if ( $a == $g ) return false;
	return true;
}
function netaddr_check($adr, $msk) {
	$a=ip2long($adr);
	if (strpos($msk,'.')!==false) {
		$m=ip2long($msk);
	} else {
		$m=pow(2, 32)-pow(2, 32-$msk);
	}
	
	if ( ($a & $m) == $a ) return true;
	return false;
	
}

function has_shell_metachars($str) {
     return preg_match('/[][\\$*?|&;()<>"\*\?\'\\\\]/', $str);
}

function save_file($name, $data, $secret=false, $group='raperca', $ignore_empty=true) {
	if ($data=="" && $ignore_empty) return ;

	$fname=tempnam(dirname($name), basename($name));
	if (file_put_contents($fname,$data)===FALSE){
		echo "<h3>Cannot save settings</h3> File: ($name).<br/>Data: $data\n";
		return false;
	}else{
		if (!$secret) {
			chmod($fname, 0644);
		} else {
			chmod($fname, $group ? 0640 : 0600);
			if ($group)
			     chgrp($fname, $group);
		}
		exec('sync');
		if (rename($fname, $name)){
			exec('sync');
			return true;
		} else {
			unlink($fname);
			echo "<h3>Cannot save settings</h3> File: ($name).<br/>Data: $data\n";
			return false;
		}
		
	}
}
function save_symlink($target, $link) {
    if ( ! symlink($target, $link) ) {
	unlink($link);
	if ( ! symlink($target, $link) ) {
	    echo "<h3>Cannot save settings</h3> File: $link as symlink to ($target)\n";
	    return false;
	}
    }
    exec('sync');
    return true;
}

function reconfig_player() {
     $pid = trim(file_get_contents('/var/run/raperca.pid'));
     posix_kill($pid, SIGHUP);
}

function reload_mdns() {
     exec('/etc/init.d/raperca-mdns reload');
}

class ipconfig{
	const NET_ETHERNET = "ethernet";
	const NET_MODEM = "modem";
	var $netiface=self::NET_ETHERNET;
	var $dhcp=1;
	var $address="";
	var $netmask="";
	var $gateway="";
	var $hostname="";

	var $startTag="# START-spxsysconfig DO NOT REMOVE THE START AND END TAGS";
	var $endTag="# END-spxsysconfig";
	var $fname="/etc/network/interfaces";
	function mask($adr,$msk){
		$a=ip2long($adr);
		$m=ip2long($msk);
		return long2ip($a & $m);
	}
	function broad($adr,$msk){
		$a=ip2long($adr);
		$m=ip2long($msk);
		
		return long2ip(($a & $m)+(~$m));
	}

	function load(){
		 // get the config
		$interface=file_get_contents($this->fname);
		if (!isset($interface) && $interface) return;
	
		$start=strpos($interface,$this->startTag);
		$end=strpos($interface,$this->endTag);
		if ($start===false || $end===false) return;

		// should look for first auto interface to be
		// consistent with ifwatchdog
		$str1="auto ";
		$iface=strpos($interface,$str1, $start);
		$iface+=strlen($str1);
		$line_end=strpos($interface,"\n",$iface);
		$ifname = substr($interface,$iface,$line_end-$iface);

		if ($ifname == "ppp0") 
			$this->netiface = self::NET_MODEM;
		else
			$this->netiface = self::NET_ETHERNET;

		$str1="iface $ifname inet ";
		$iface=strpos($interface,$str1, $start);
		$iface+=strlen($str1);
		$conf=substr($interface,$iface,$end-$iface);
		$vals=explode ("\n",$conf);	
		
		if (trim($vals[0])=="dhcp"){
			$this->dhcp=1;
			foreach ($vals as $name){
				if (!(($n=strpos($name,"hostname "))===false)){
					$this->hostname=trim(substr($name, strlen("hostname ")));
				}
			}
		}elseif (trim($vals[0])=="static"){
			$this->dhcp=0;
			foreach ($vals as $name){
				if (!(($n=strpos($name,"address "))===false)){
					$this->address=trim(substr($name, strlen("address")+$n));
				}elseif (!(($n=strpos($name,"netmask "))===false)){
					$this->netmask=trim(substr($name, strlen("netmask")+$n));
				}elseif (!(($n=strpos($name,"gateway "))===false)){
					$this->gateway=trim(substr($name, strlen("gateway")+$n));
				}
			}
		}elseif (trim($vals[0])=="ppp"){
			// nothing else to read, we always use the same ppp config
		}
		
	}
	function save(){
		if ($this->netiface == self::NET_MODEM)
			$iface="ppp0";
		else
			$iface="eth0";
		$conf="auto $iface\niface $iface inet ";
		$hasError=false;
		$hasGateway=true;
		if ($this->netiface == self::NET_MODEM) {
			$conf.="ppp\n";
			$conf.="\tprovider modem\n";
			$conf.="\tpre-up /usr/libexec/spxdevs/modem-check\n";
		} else if (!$this->dhcp){
		     if (!ip_check($this->address)){ 
		     	echo "<p style='color:red'>Warning: '".htmlspecialchars($this->address)."' is not a valid address. Configuration not saved</p>"; 
		     	$hasError=true;
	     	} 
		     if (!netmask_check($this->netmask)){ 
		     	echo "<p style='color:red'>Warning: '".htmlspecialchars($this->netmask)."' is not a valid netmask. Configuration not saved</p>"; 
		     	$hasError=true;
	     	}
		     if ($this->gateway!=""){
		     	if (!ip_check($this->gateway)){ 
		     		echo "<p style='color:red'>Warning: ".htmlspecialchars($this->gateway)." is not a valid gateway. Configuration not saved</p>"; 
		     		$hasError=true;
		     	}
		     	if (!ipconf_check($this->address, $this->netmask, $this->gateway)){ 
		     		echo "<p style='color:red'>Warning: IP configuration is not valid. Configuration not saved</p>"; 
		     		$hasError=true;
		     	}
		     } else $hasGateway=false;
			
			if ($hasError) return false;
			
			$conf.="static\n";
			$conf.="\taddress $this->address\n";
			$conf.="\tnetwork ".$this->mask($this->address,$this->netmask)."\n";
			$conf.="\tnetmask $this->netmask\n";
			if ($hasGateway) $conf.="\tgateway $this->gateway\n";
			$conf.="\tbroadcast ".$this->broad($this->address,$this->netmask)."\n";
			$conf.="\tpre-up [ ! -x /etc/network/eth-inet-dad ] || /etc/network/eth-inet-dad\n";
			
		} else {
			$conf.="dhcp\n";
			if ($this->hostname!="")$conf.="\thostname $this->hostname\n";
		}
			

		$interface=file_get_contents($this->fname);
		if (!isset($interface) && $interface) return false;
	
		$start=strpos($interface,$this->startTag);
		$end=strpos($interface,$this->endTag);
		if ($start===false) $start=strlen($interface);	
		if ($end===false) {
			$end=$start;	
			$conf.=$this->endTag."\n";
		}
		if (!save_file($this->fname,substr_replace($interface,$this->startTag."\n".$conf,$start,$end-$start)))
			return false;

		// copy settings to failsafe-data
		if ($this->netiface == self::NET_MODEM) {
			$netconf = "auto\n";
		} elseif ($this->dhcp) {
		    $netconf = "dhcp\n";
		} else {
		    $netconf = "static=" . $this->address . ";" .
			$this->netmask . ";" .
			$this->broad($this->address,$this->netmask ) . ";";
		    if ($hasGateway)
			$netconf .= $this->gateway;
		    $netconf .= "\n";
		}

		if ( ! setSafeModeData("network", $netconf, FALSE) )
		    echo "<p style='color:red'>Failed saving network configuration for recovery console, please try again.</p>";

		return true;
	}
}

class dnsconfig{
	var $nameserver1="";
	var $nameserver2="";
	var $nameserver3="";
	var $search="";
	var $domain="";
	var $fname="/etc/network/resolv.conf.static";
	function load(){
		$dns=file_get_contents($this->fname);
		$tmp=$dns;
		$i=1;
		$pos=strpos($tmp,"nameserver");
		while (!($pos===false)){
			$pos+=strlen("nameserver");
			$name="nameserver".$i; 
			$i++;
			$this->$name=trim(substr($tmp,$pos,strpos($tmp,"\n",$pos)-$pos));
			$tmp=substr($tmp,strpos($tmp,"\n",$pos));
			$pos=strpos($tmp,"nameserver");
		}
			
		if (!($pos===false)) 
		$pos=strpos($dns,"search ");
		if (!($pos===false)) {$pos+=strlen("search "); $this->search=trim(substr($dns,$pos,strpos($dns,"\n",$pos)-$pos));}
		$pos=strpos($dns,"domain ");
		if (!($pos===false)) {$pos+=strlen("domain "); $this->domain=trim(substr($dns,$pos,strpos($dns,"\n",$pos)-$pos));}
	}
	function save(){
		$conf="";
		if ($this->nameserver1!="" ) 
			if (ip_check($this->nameserver1)) $conf.="nameserver $this->nameserver1\n"; 
			else { echo "<p style='color:red'>Warning: ".htmlspecialchars($this->nameserver1)." is not a valid DNS. Configuration not saved</p>"; return false;}
		if ($this->nameserver2!="" ) 
			if (ip_check($this->nameserver2)) $conf.="nameserver $this->nameserver2\n"; 
			else { echo "<p style='color:red'>Warning: ".htmlspecialchars($this->nameserver2)." is not a valid DNS. Configuration not saved</p>"; return false;}
		if ($this->nameserver3!="" ) 
			if (ip_check($this->nameserver3)) $conf.="nameserver $this->nameserver3\n"; 
			else { echo "<p style='color:red'>Warning: ".htmlspecialchars($this->nameserver3)." is not a valid DNS. Configuration not saved</p>"; return false;}
		
		if ($this->search!="") $conf.="search $this->search\n";
		if ($this->domain!="") $conf.="domain $this->domain\n";
		if ( $conf == "" ) $conf="\n"; 
		return save_file($this->fname,$conf);
	}
	function clear(){
	     $this->nameserver1="";
	     $this->nameserver2="";
	     $this->nameserver3="";
	     $this->search="";
	     $this->domain="";
	}
}

class modemconfig {
    // NOTE: The 3GPP AT commands are defined in 3GPP TS 27.007

    const NUM_INIT = 6;
    private static $form_prefix = 'm_';
    private static $fdesc = array(
        "pin" => "PIN",
        "puk" => "PUK",
        "dial" => "dial number",
        "apn" => "APN",
        "pdptype" => "PDP type",
        "username" => "username",
        "password" => "password",
        "authproto" => "auth protocol");
    private static $defaults = array(
        // init1 to init3 are the defaults used in simple mode
        "init1" => "AT Z", // after ATZ there can be no trailing commands
        "init2" => "AT Q0 V1 E1 &C1 &D2 +CMEE=2",
        "init3" => "AT +CGREG=2; +CR=1; +CGREG?; +CGEREP=1",
        "dial" => "*99**1#",
        "pdptype" => "IP",
        "authproto" => "any");
    private static $choice = array(
        "pdptype" => array('IP', 'PPP'),
        "authproto" => array('any', 'none', 'pap', 'chap')
        );
    private static $type = array( // if not in the table, this is advanced
        "pin" => "all",
        "puk" => "all",
		"pdptype" => "simple",
        "apn" => "simple",
		"username" => "all",
        "password" => "all",
        "authproto" => "all"
        );
    private static $help = array(
        'puk' => "required to reset the PIN to the above value",
        'init1' => "e.g., '&nbsp;AT Z&nbsp;'",
        "init2" => "e.g., '&nbsp;AT Q0 V1 E1 &amp;C1 &amp;D2 +CMEE=2&nbsp;'",
        "init3" => "e.g., '&nbsp;AT +CGREG=2; +CR=1; +CGREG?; +CGEREP=1&nbsp;'",
        "init4" => "e.g., '&nbsp;AT +CGDCONT=<i>&lt;cid&gt;</i>,\"IP\",\"<i>&lt;apn&gt;</i>\"&nbsp;'",
        "dial" => "e.g., '&nbsp;*99**1#&nbsp;' or '&nbsp;*99**1*<i>&lt;cid&gt;</i>#&nbsp;'",
        "username" => "use [serial] as placeholder for the serial number" );
    private static $peer_file = '/etc/ppp/peers/modem';
    private static $chat_file = '/etc/chatscripts/modem';
    private static $conf_file = '/etc/network/modem.conf';
    private static $pass_file = '/etc/ppp/modem-password';
    private static $pin_file = '/etc/chatscripts/modem-pin';
    private static $modem_ctl = '/usr/libexec/spxdevs/modem-control';
    private static $status_dir = '/dev/.modem-data';
    private static $devnode = '/dev/modem';
    private static $devnodectl = '/dev/modemctl';

    private $fields = array();
    private $changed;
    private $status = '';
    private $connect_status = '';
    private $simple;

    static function getFormNames() {
        static $fnames;
        if (isset($fnames))
            return $fnames;
        $fnames[] = self::$form_prefix . 'pin';
        $fnames[] = self::$form_prefix . 'puk';
        for ($i = 1; $i <= self::NUM_INIT; $i++) 
            $fnames[] = self::$form_prefix . "init$i";
        foreach (self::$fdesc as $key => $val) {
            if ($key != 'pin' && $key != 'puk')
                $fnames[] = self::$form_prefix . "$key";
        }
        return $fnames;
    }

    static function getFieldDesc($name) {
        $name = preg_replace("/^" . self::$form_prefix . "/", '', $name);
        if (strncmp($name, 'init', 4) == 0)
            return 'init string #' . substr($name, 4);
        else
            return self::$fdesc[$name];
    }

    static function getFieldHelp($name) {
        $name = preg_replace("/^" . self::$form_prefix . "/", '', $name);
        if (array_key_exists($name, self::$help))
            return self::$help[$name];
        else
            return null;
    }
    function isSimple() {
    	return $this->simple;
    }
    function setSimple ( $isSimple ){
    	$this->simple = $isSimple;
    }
    static function isSimpleOnly($name) {
        $name = preg_replace("/^" . self::$form_prefix . "/", '', $name);
        return array_key_exists($name, self::$type) &&  self::$type[$name]=="simple";
    }
    static function isAdvancedOnly($name) {
        $name = preg_replace("/^" . self::$form_prefix . "/", '', $name);
        return !array_key_exists($name, self::$type);
    }

    static function isChoiceField($name) {
        $name = preg_replace("/^" . self::$form_prefix . "/", '', $name);
        return array_key_exists($name, self::$choice);
    }

    static function getFieldChoices($name) {
        $name = preg_replace("/^" . self::$form_prefix . "/", '', $name);
        return self::$choice[$name];
    }

    /**
     * Returns the signal quality. This method is slow, so do not use
     * int time critical places.
     * @return The signal quality as a percentage or FALSE
     * on error
     */
    static function getSignalQuality() {
        if ( ! file_exists(self::$devnodectl) )
            return false;
        $cmd = self::$modem_ctl .
            " -c " . escapeshellarg('+CSQ') .
            " " . self::$devnodectl;
        exec($cmd, $out, $retval);
        if ($retval != 0)
            return false;
        foreach ($out as $line) {
            if (preg_match("@^\+CSQ:\s+(\d+)@", $line, $matches)) {
                $rssi = $matches[1];
                if ($rssi < 0 || $rssi >= 99)
                    return false; // unknown or not detectable
                $dbm = -113 + 2 * $rssi;
                if ($dbm >= -65)
                    return 1;
                return round(100 * ($dbm+113) / 48);
            }
        }
        return false;
    }

    static function quote($str, $enclose = true) {
        static $from = array("\\", "\"");
        static $to = array("\\\\", "\\\"");
        $str = str_replace($from, $to, $str);
        if ( $enclose )
            return '"' . $str . '"';
        else
            return str_replace(' ', '\s', $str); // escape spaces
    }

    static function atstring($str) {
        static $from = array("\\", "\"");
        static $to = array("\\5C", "\\22");
        return '"' . str_replace($from, $to, $str) . '"';
    }

    function isFieldActive($name) {
        $name = preg_replace("/^" . self::$form_prefix . "/", '', $name);
        if ($name == 'puk')
            return $this->isPINBlocked();
        return true;
    }

    function setField($name, $val) {
        $name = preg_replace("/^" . self::$form_prefix . "/", '', $name);
        $val = isset($val) ? trim($val) : $val='';
        if ($this->simple) {
            if (!array_key_exists($name, self::$type))
                $val=''; // advanced only field
        } else {
            if (array_key_exists($name, self::$type) && self::$type[$name]=="simple")
                $val=''; // simple only field
        }
        if (array_key_exists($name, $this->fields)) {
            if ($this->fields[$name] == $val)
                return;
        } else {
            if (strlen($val) == 0)
                return;
        }
        if (strlen($val) != 0)
            $this->fields[$name] = $val;
        else
            unset($this->fields[$name]);
        $this->changed = true;
    }

    function getField($name) {
        $name = preg_replace("/^" . self::$form_prefix . "/", '', $name);
        if (array_key_exists($name, $this->fields))
            return $this->fields[$name];
        else
            return '';
    }

    function isPINBlocked() {
        return ($this->getStatus() == 'PUK' || $this->getStatus() == 'PUK INCORRECT');
    }

    function isChanged() {
        return $this->changed;
    }

	function __construct() {
		$this->clear();
	}

	function load() {
        $xml = new DOMDocument();
        if ( ! $xml->load(self::$conf_file) )
            return;
        $ml = $xml->getElementsByTagName('modem');
        if ($ml->length == 0)
            return;
        $this->fields = array();
        $modem = $ml->item(0);
        $n_init = 0;
        foreach ($ml->item(0)->childNodes as $item) {
            if ($item->nodeType != XML_ELEMENT_NODE)
                continue; // skip all comments, whitespace, etc.
            $name = $item->nodeName;
            if (strncmp($name, 'init', 4) == 0) {
                $n_init++;
                $name = $name . $n_init;
            }
            $this->fields[$name] = trim($item->textContent);
        }
        $pin = file_get_contents(self::$pin_file);
        if ($pin)
            $this->fields['pin'] = $pin;
        $pass = file_get_contents(self::$pass_file);
        if ($pass)
            $this->fields['password'] = $pass;
        $this->changed = false;

        // The prescence of pdptype means a simple configuration
        $this->simple = array_key_exists('pdptype', $this->fields);
	}

	function save() {
        if ( ! $this->check() )
            return false;
            
		if ( $this->simple ){ // simple config, set implicit parameters
        	$this->fields['init1'] = self::$defaults['init1'];
        	$this->fields['init2'] = self::$defaults['init2'];
        	$this->fields['init3'] = self::$defaults['init3'];
        	$this->fields['init4'] =
                'AT +CGDCONT=1,' . $this->atstring($this->fields["pdptype"]) .
                ',' . $this->atstring($this->fields["apn"]);
        	$this->fields['dial'] = '*99**1*1#';
        }
        
        // create the xml modem config
        $xml = new DOMDocument();
        $xml->formatOutput = true;
        $root = $xml->createElement('modems');
        $xml->appendChild($root);
        $modem = $xml->createElement('modem');
        $root->appendChild($modem);
        foreach ($this->fields as $name => $field) {
            if (strncmp($name, 'init', 4) == 0) {
                $xname = 'init';
            } elseif ($name == 'pin') {
                continue;
            } elseif ($name == 'puk') {
                continue;
            } elseif ($name == 'password') {
                continue;
            } else {
                $xname = $name;
            }
            $node = $xml->createElement($xname);
            $modem->appendChild($node);
            $node->appendChild($xml->createTextNode($field));
        }
        

        // create chat script
        $chat = '';
        $chat .= "# Use a small timeout while we set everything up\n";
        $chat .= "TIMEOUT 3\n";
        $chat .= "# The network failue strings on which we abort\n";
        $chat .= "# we timeout instead of abort on error messages\n";
        $chat .= "# so as to get the full message in the logs\n";
        $chat .= "ABORT BUSY\n";
        $chat .= "ABORT 'NO CARRIER'\n";
        $chat .= "ABORT VOICE\n";
        $chat .= "ABORT 'NO DIALTONE'\n";
        $chat .= "ABORT 'NO DIAL TONE'\n";
        $chat .= "ABORT 'NO ANSWER'\n";
        $chat .= "ABORT DELAYED\n";
        $noexpect = true;
        $chat .= "# --- Initialization\n";
        $chat .= "SAY \"--- passing init strings ---\\n\"\n";
        for ($i = 1; $i <= self::NUM_INIT; $i++) {
            $name = "init$i";
            if (array_key_exists($name, $this->fields)) {
                $chat .= $noexpect ? "'' " : "OK\\r\\n ";
                $chat .= $this->quote($this->fields[$name]) . "\n";
                $noexpect = false;
            }
        }
        if ( ! $noexpect ) {
            $chat .= "OK\\r\\n \\c\n";
            $noexpect = true;
        }
        if (array_key_exists('dial', $this->fields)) {
            $chat .= "# --- Dial\n";
            $chat .= $noexpect ? "'' " : "OK\\r\\n ";
            $chat .= $this->quote("ATD{$this->fields['dial']}") . "\n";
            $noexpect = false;
        }
        $chat .= "# --- Done, expect data state confirmation and transition to data state\n";
        $chat .= "# Session establishment may take a long time\n";
        $chat .= "TIMEOUT 20\n";
        $chat .= "SAY \"--- waiting for connection ---\\n\"\n";
        $chat .= "CONNECT \c\n";
        $chat .= "\\n \\c\n";
        $chat .= "SAY \"--- entered data state ---\\n\"\n";

        // create ppp peer definition
        $peer = '';
        $peer .= "# Enable debugging\n";
        $peer .= "debug\n";
        $peer .= "# Do not detach until the protocol is up but discard logs on stdout\n";
        $peer .= "updetach\n";
        $peer .= "logfile /dev/null\n";
        $peer .= "# Be sure to use ppp0 as the interface\n";
        $peer .= "unit 0\n";
        $peer .= "# The modem device file\n";
        $peer .= self::$devnode . "\n";
        $peer .= "# Create the lock file for the device\n";
        $peer .= "lock\n";
        $peer .= "# Use hardware flow control\n";
        $peer .= "crtscts\n";
        $peer .= "# Be sure ppp makes use of control lines\n";
        $peer .= "modem\n";
        $peer .= "# Use pid file /var/run/ppp-modem.pid\n";
        $peer .= "linkname modem\n";
        $peer .= "# Do not require peer to authenticate\n";
        $peer .= "noauth\n";
        $peer .= "# Do not attempt to determine local IP address from hostname\n";
        $peer .= "noipdefault\n";
        $peer .= "# Do not let ppp enter a default route; it is handled elsewhere\n";
        $peer .= "nodefaultroute\n";
        $peer .= "# Get DNS servers from peer\n";
        $peer .= "usepeerdns\n";
        $peer .= "# Do not do proxy ARP\n";
        $peer .= "noproxyarp\n";
        $peer .= "# Do not retry after connection goes down, this is handled elsewhere\n";
        $peer .= "nopersist\n";
        if ( $this->simple && strncmp($this->fields["pdptype"], 'IP', 2) == 0 ) {
            $peer .= "# Modem using native IP connection, disable PPP data compression\n";
            $peer .= "nobsdcomp\n";
            $peer .= "nodeflate\n";
            $peer .= "nopredictor1\n";
            $peer .= "# Modem using native IP connection, disable IP header compression\n";
            $peer .= "novj\n";
        }
        $peer .= "# The script to initialize the modem connection\n";
        $peer .= "connect /usr/libexec/spxdevs/modem-connect\n";
        $peer .= "# The script to bring down the modem connection\n";
        $peer .= "disconnect \"/usr/libexec/spxdevs/modem-connect -d\"\n";
        if (array_key_exists('username', $this->fields)) {
            global $serial;
            $username = str_replace('[serial]', $serial, $this->fields['username']);
            $peer .= "# The user to use in authentication\n";
            $peer .= "user " . $this->quote($username) . "\n";
        }
        if (array_key_exists('authproto', $this->fields)) {
            switch ($this->fields['authproto']) {
            case 'chap':
                $peer .= "# Use 'modem-chap' as remote name for CHAP only auth\n";
                $peer .= "remotename modem-chap\n";
                break;
            case 'pap':
                $peer .= "# Use 'modem-pap' as remote name for PAP only auth\n";
                $peer .= "remotename modem-pap\n";
                break;
            case 'none':
                $peer .= "# Use 'modem-none' as remote name for no authentication\n";
                $peer .= "remotename modem-none\n";
                break;
            case 'any':
                $peer .= "# Use 'modem' as remote name for auth with any protocol\n";
                $peer .= "remotename modem\n";
                break;
            }
        }

        // The pin password values
        $pin = array_key_exists('pin', $this->fields) ? $this->fields['pin'] : '';
        $pass = array_key_exists('password', $this->fields) ? $this->fields['password'] : '';

        $allOK = save_file(self::$peer_file, $peer);
        if ($allOK)
            $allOK = save_file(self::$chat_file, $chat);
        if ($allOK)
            $allOK = save_file(self::$conf_file, $xml->saveXML());
        if ($allOK)
            $allOK = save_file(self::$pin_file, $pin, true, null, false);
        if ($allOK)
            $allOK = save_file(self::$pass_file, $pass, true, null, false);

		return $allOK;
	}

	function check() {
		$isOK=true;
        $n_modem_cmds = 0;

        // check for valid string (only ASCII non-control characters)
        foreach ($this->fields as $name => $field) {
            if (strlen($field) != 0 && preg_match("/[^\x20-\x7E]/", $field)) {
                echo "<p style='color:red'>Warning: " . htmlspecialchars($field) .
                    " is not a valid " . htmlspecialchars(self::getFieldDesc($name)) .
                    ". Configuration not saved.</p>";
                $isOK=false;
            }
        }

        // check that dial string has no ';' except as last character
        if (array_key_exists('dial', $this->fields)) {
            $n_modem_cmds++;
            $dial = $this->fields['dial'];
            $p = strpos($dial, ';');
            if ($p !== FALSE && ($p != strlen($dial)-1 || strlen($dial) == 1)) {
                echo "<p style='color:red'>Warning: the " .
                    htmlspecialchars(self::getFieldDesc('dial')) .
                    " can only have a ';' at the end. Configuration not saved.</p>";
                $isOK = false;
            }
        }

        // check that PDP type is present in simple mode
        if ($this->simple && !array_key_exists('pdptype', $this->fields)) {
            echo "<p style='color:red'>Warning: the " .
                htmlspecialchars(self::getFieldDesc('pdptype')) .
                " is required in simple mode. Configuration not saved.</p>";
            $isOK = false;
        }

        // check that PDP type, if any, is valid
        if (array_key_exists('pdptype', $this->fields)) {
            $n_modem_cmds++; // a PDP type generates a modem command
            $pdptype = $this->fields['pdptype'];
            // Other valid types would be IPV6 and IPV4V6 but we have not
            // verified IPv6 support, so we do not accept them yet
            if ($pdptype != 'IP' && $pdptype != 'PPP') {
                echo "<p style='color:red'>Warning: the " .
                    htmlspecialchars(self::getFieldDesc('pdptype')) .
                    " must be IP or PPP. Configuration not saved.</p>";
                $isOK = false;
            }
        }

		// check that each init string starts with AT
        for ($i = 1; $i <= self::NUM_INIT; $i++) {
            $name = "init$i";
            if (array_key_exists($name, $this->fields)) {
                $n_modem_cmds++;
                $field = $this->fields[$name];
                if (substr($field, 0, 2) != 'AT' && substr($field, 0, 2) != 'at' ) {
                    echo "<p style='color:red'>Warning: init strings must start with 'AT' or 'at'. " .
                        "Configuration not saved.</p>";
                    $isOK = false;
                }
            }
        }

        // check that there is at least one modem command (besides PIN)
        if ($n_modem_cmds <= 0) {
            echo "<p style='color:red'>Warning: at least one init string or dial number " .
                "must be entered. Configuration not saved.</p>";
            $isOK = false;
        }

        // check for valid authentication protocols
        if (array_key_exists('authproto', $this->fields)) {
            if ( ! in_array($this->fields['authproto'], self::$choice['authproto'])) {
                    echo "<p style='color:red'>Warning: invalid authentication protocol. " .
                        "Configuration not saved.</p>";
                    $isOK = false;
            }
        }
		return $isOK;
	}

	function clear() {
        $this->fields = self::$defaults;
        $this->changed = true;
        $this->simple = true;
	}

    function getStatus($refresh = false)
    {
        if ( $refresh || ! $this->status ) {
            if ( ! file_exists(self::$devnode) ||
                 ! file_exists(self::$status_dir . '/status') )
            {
                $this->status = 'NOT FOUND';
            } else {
                $this->status = trim(file_get_contents(self::$status_dir . '/status'));
            }
        }
        return $this->status;
    }

    function getConnectStatus($refresh = false)
    {
        if ( $refresh || ! $this->connect_status ) {
            if ( ! file_exists(self::$status_dir . '/connect-status') )
            {
                $this->connect_status = '';
            } else {
                $this->connect_status =
                    trim(file_get_contents(self::$status_dir . '/connect-status'));
            }
        }
        return $this->connect_status;
    }

    function resetPIN()
    {
        if ( ! $this->isPINBlocked() ) {
            echo "<p style='color:red'>Warning: attempted to reset PIN when not blocked. " .
                "Configuration not saved.</p>";
            return false;
        }
        $puk = $this->getField('puk');
        $newpin = $this->getField('pin');
        if ( ! $puk ) {
            echo "<p style='color:red'>Warning: the PUK is required to reset the PIN. " .
                "Configuration not saved.</p>";
            return false;
        }
        if ( ! file_exists(self::$devnodectl) ) {
            echo "<p style='color:red'>Warning: no modem control device found to reset the PIN. " .
                "Configuration not saved.</p>";
            return false;
        }
        $cmd = self::$modem_ctl .
            " -p " . escapeshellarg($newpin) .
            " -P " . escapeshellarg($puk) .
            " " . self::$devnodectl;
        exec($cmd, $out, $retval);
        $status = $this->getStatus(true); // refreshes the status value
        if ($retval != 0) {
            echo "<p style='color:red'>Warning: failed resetting the PIN";
            if ($status == 'PUK INCORRECT')
                echo " (PUK was incorrect)";
            echo ". Configuration not saved.</p>";
            unset($this->fields['puk']); // so that we do not show previous PUK in form
            return false;
        }
        return true;
    }
}

function shutdown($msg = null, $forcereboot=false) {
	global $installationmode;
	global $product;
	if ($installationmode && !$forcereboot){
		if ( $msg == null ) $msg="unknown reason";
		if (file_exists("/var/run/reboot-pending"))
			$pending = file_get_contents("/var/run/reboot-pending");
		else $pending="";
		$pending = $pending.$msg."\n";
		file_put_contents("/var/run/reboot-pending", $pending);
		echo "<h3>Change saved.</h3><p>Adding the following to the pending change list:</p>\n";
		echo "<ul><li>$msg</li></ul>\n";
		echo "<p>Make sure you restart the $product to activate it.</p>\n";
		syslog(LOG_NOTICE, "configuration change in installation mode:" . $msg);
	} else {
	     echo "<h3>Device is restarting.</h3>\n";
	     echo "<p>There will be no indication that the $product is back online. Please wait for 2 minutes before using the http interface again.</p>\n";
		if ($installationmode)
			syslog(LOG_NOTICE, "installation mode shutdown" . ($msg ? (": " . $msg) : ''));
		else
			syslog(LOG_NOTICE, "user initiated shutdown" . ($msg ? (": " . $msg) : ''));
		exec('/sbin/shutdown -r now');
	}
}

function recovery_version() {
	
	$info = file_get_contents("/proc/mtd");
	
	foreach (explode("\n", $info) as $str ){
		if ( strpos($str, "failsafe-fs")===false ) continue;
		$max = strpos($str, ":");
		$mtd = substr($str, 0, $max);
		
		$f = fopen("/dev/".$mtd."ro", 'r');
		if ($f===false) return "[ro]";
		$data = fread($f, 64);
		if ($data===false) return "[read]";
		
		if (ord(substr($data,0,1))!=0x27 || ord(substr($data,1,1))!=0x05 || ord(substr($data,2,1))!=0x19 || ord(substr($data,3,1))!=0x56 ) return "[magic]";
		$values = explode(" ", substr($data, 32));
		if (count($values)==0) return "[data]";
		return $values[count($values)-1];
	}
	return "[not found]";
}

class ComPort {
	
	var $confpathsys="/usr/share/raperca/comportfsm/";
	var $confpathusr="/srv/raperca/comportfsm/";
	var $powersavefile="power_save.fsm";
	
	var $monPower="false";
	var $onTime="00:00";
	var $offTime="24:00";
	var $powerSave="true";
	var $href="";
	var $portNumber=1;
	
	var $comSettings="false";
	var $baudRate=9600;
	var $byteSize=8;
	var $stopBits=1;
	var $parity="none";
	var $flowControl="none";
	
	var $modified = false;
	
	function set( $inputs ) {

		$values = array("onTime", "offTime", "powerSave", "portNumber", "baudRate", "byteSize", "stopBits", "parity", "comSettings", "monPower");
		foreach ($values as $str )
			if ( isset( $inputs[$str] ) && (!isset($this->$str) || $inputs[$str]!=$this->$str) ){
				if ( isset($this->$str) )
					$this->modified=true;
				$this->$str = $inputs[$str];
			}
			if ( isset( $inputs["href"] ) && $inputs["href"]!=$this->href ){
				if ( ($this->href!="" && $this->href!="[disabled]" ) || ($inputs["href"]!="" && $inputs["href"]!="[disabled]") ){
					$this->modified=true;
					$this->href = $inputs["href"];
				}
			}

		return $this->modified;
	}
	
	function load( $domtree ) {
		$ComPorts=$domtree->getElementsByTagName("serialPort");
		if ($ComPorts->length>0){
			$ComPorts=$ComPorts->item(0);
			
			if ($ComPorts->hasAttribute("enablePowerSave"))
				$this->powerSave=$ComPorts->getAttribute("enablePowerSave"); 
			if ($ComPorts->hasAttribute("href")) 
				$this->href=$ComPorts->getAttribute("href");
			
			if ( $this->href=="" ) {
				$this->href = "[disabled]";
				$this->powerSave="false";
			}elseif ($this->href==$this->confpathsys.$this->powersavefile){
				$this->href="[disabled]";
				$this->powerSave="true";
			}
			
			if ($ComPorts->hasAttribute("onTime") && $ComPorts->hasAttribute("offTime")) {
				$this->onTime=$ComPorts->getAttribute("onTime");
				$this->offTime=$ComPorts->getAttribute("offTime");
			}
			if ( $this->onTime!="00:00" || $this->offTime!="24:00" ){
				$this->monPower="true";
			}
			
			
			if ($ComPorts->hasAttribute("portNumber")) $this->portNumber=$ComPorts->getAttribute("portNumber");
			
			if ($ComPorts->hasAttribute("baudRate")) { $this->baudRate=$ComPorts->getAttribute("baudRate"); $this->comSettings="true";}
			if ($ComPorts->hasAttribute("byteSize")) { $this->byteSize=$ComPorts->getAttribute("byteSize"); $this->comSettings="true";}
			if ($ComPorts->hasAttribute("stopBits")) { $this->stopBits=$ComPorts->getAttribute("stopBits"); $this->comSettings="true";}
			if ($ComPorts->hasAttribute("flowControl")) { $this->flowControl=$ComPorts->getAttribute("flowControl"); $this->comSettings="true";}
			if ($ComPorts->hasAttribute("parity")) { $this->parity=$ComPorts->getAttribute("parity"); $this->comSettings="true";}
			
			
		} else {
			$this->href = "[disabled]";
			$this->powerSave="false";
		}
	}
	
	function save( $domtree ) {
		$ComPorts=$domtree->getElementsByTagName("serialPort");
		
		if ( $this->href=="[disabled]")
			$this->href="";
		if (  $this->powerSave=="true" && $this->href==""){
			$this->href = $this->confpathsys.$this->powersavefile;
		}
		if ( $this->href!="" ) {
			if ($ComPorts->length>0){
				$ComPorts=$ComPorts->item(0);
			} else {
				$ComPorts=$domtree->createElement("serialPort");	
				$player=$domtree->documentElement->getElementsByTagName("player");
				if ($player->length>0) {
					$ComPorts=$player->item(0)->appendChild($ComPorts);
				}else{
					$ComPorts=$domtree->documentElement->appendChild($ComPorts);
					echo "Warning no player tag found in the config, cannot add SerialPort tag.<br/>";
				}
			}
			
			$ComPorts->setAttribute( "href", $this->href );
			if ( $this->monPower=="true" && ($this->onTime!="00:00" || $this->offTime!="24:00")){
				$ComPorts->setAttribute( "onTime", $this->onTime );
				$ComPorts->setAttribute( "offTime", $this->offTime );
			} else {
				$this->monPower="false";
				$ComPorts->removeAttribute("onTime");
				$ComPorts->removeAttribute("offTime");
			}
			$ComPorts->setAttribute( "enablePowerSave", $this->powerSave );
			
			$ComPorts->setAttribute("portNumber",$this->portNumber);
			if ( $this->comSettings=="true" ){
				$ComPorts->setAttribute("baudRate",$this->baudRate);
				$ComPorts->setAttribute("byteSize",$this->byteSize);
				$ComPorts->setAttribute("stopBits",$this->stopBits);
				$ComPorts->setAttribute("flowControl",$this->flowControl);
				$ComPorts->setAttribute("parity",$this->parity);
			} else {
				$ComPorts->removeAttribute("baudRate");
				$ComPorts->removeAttribute("byteSize");
				$ComPorts->removeAttribute("stopBits");
				$ComPorts->removeAttribute("flowControl");
				$ComPorts->removeAttribute("parity");
			}
		} else {
			if ($ComPorts->length>=0){
				$deleteList=array();
				foreach($ComPorts as $item) $deleteList[]=$item;
				foreach($deleteList as $item) $item->parentNode->removeChild($item);
			}
		}
		return $this->modified;
	}
	
	function getFileList() {
		$filelist1=scandir($this->confpathsys);
		$filelist=array("[disabled]" => "[disabled]");
		foreach ($filelist1 as $file){
			if ($file=="." || $file==".." || $file==$this->powersavefile) continue;
			$filelist["[sys] ".basename($file,".fsm")]=$this->confpathsys.$file;
		}
	    $filelist2=scandir($this->confpathusr);
	    foreach ($filelist2 as $file){
			if ($file!="." && $file!="..") $filelist[basename($file,".fsm")]=$this->confpathusr.$file;
	    }
		return $filelist;
	}
	
	function addFile( $filename ) {
		if (array_key_exists($filename,$_FILES)){
			if ($_FILES[$filename]["error"]>0 && $_FILES[$filename]["name"]){
			     echo "Error: ", htmlspecialchars($_FILES[$filename]["error"]), " while uploading: ", htmlspecialchars($_FILES[$filename]["name"]), "<br />";
			}else {
				$ext=pathinfo($_FILES[$filename]["name"]);
				if (strcasecmp("fsm",$ext["extension"])==0){
					move_uploaded_file($_FILES[$filename]["tmp_name"],$this->confpathusr.$_FILES[$filename]["name"]);
					return true;
				}else{
				     echo "Error: ", htmlspecialchars($_FILES[$filename]["name"]), " is not an fsm file<br />";
				}
			}
	    }
	    return false;
	}
}
